@@ -19,8 +19,9 @@ from mock import call
 from osc_lib import exceptions
 from osc_lib import utils
 
+from openstackclient.tests.unit.identity.v3 import fakes as project_fakes
 from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
-from openstackclient.volume.v2 import snapshot
+from openstackclient.volume.v2 import volume_snapshot
 
 
 class TestSnapshot(volume_fakes.TestVolume):
@@ -32,6 +33,8 @@ class TestSnapshot(volume_fakes.TestVolume):
         self.snapshots_mock.reset_mock()
         self.volumes_mock = self.app.client_manager.volume.volumes
         self.volumes_mock.reset_mock()
+        self.project_mock = self.app.client_manager.identity.projects
+        self.project_mock.reset_mock()
 
 
 class TestSnapshotCreate(TestSnapshot):
@@ -67,24 +70,25 @@ class TestSnapshotCreate(TestSnapshot):
 
         self.volumes_mock.get.return_value = self.volume
         self.snapshots_mock.create.return_value = self.new_snapshot
+        self.snapshots_mock.manage.return_value = self.new_snapshot
         # Get the command object to test
-        self.cmd = snapshot.CreateSnapshot(self.app, None)
+        self.cmd = volume_snapshot.CreateVolumeSnapshot(self.app, None)
 
     def test_snapshot_create(self):
         arglist = [
-            "--name", self.new_snapshot.name,
+            "--volume", self.new_snapshot.volume_id,
             "--description", self.new_snapshot.description,
             "--force",
             '--property', 'Alpha=a',
             '--property', 'Beta=b',
-            self.new_snapshot.volume_id,
+            self.new_snapshot.name,
         ]
         verifylist = [
-            ("name", self.new_snapshot.name),
+            ("volume", self.new_snapshot.volume_id),
             ("description", self.new_snapshot.description),
             ("force", True),
             ('property', {'Alpha': 'a', 'Beta': 'b'}),
-            ("volume", self.new_snapshot.volume_id),
+            ("snapshot_name", self.new_snapshot.name),
         ]
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
@@ -102,7 +106,7 @@ class TestSnapshotCreate(TestSnapshot):
 
     def test_snapshot_create_without_name(self):
         arglist = [
-            self.new_snapshot.volume_id,
+            "--volume", self.new_snapshot.volume_id,
             "--description", self.new_snapshot.description,
             "--force"
         ]
@@ -125,6 +129,60 @@ class TestSnapshotCreate(TestSnapshot):
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, data)
 
+    def test_snapshot_create_without_volume(self):
+        arglist = [
+            "--description", self.new_snapshot.description,
+            "--force",
+            self.new_snapshot.name
+        ]
+        verifylist = [
+            ("description", self.new_snapshot.description),
+            ("force", True),
+            ("snapshot_name", self.new_snapshot.name)
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.volumes_mock.get.assert_called_once_with(
+            self.new_snapshot.name)
+        self.snapshots_mock.create.assert_called_once_with(
+            self.new_snapshot.volume_id,
+            force=True,
+            name=self.new_snapshot.name,
+            description=self.new_snapshot.description,
+            metadata=None,
+        )
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, data)
+
+    def test_snapshot_create_without_remote_source(self):
+        arglist = [
+            '--remote-source', 'source-name=test_source_name',
+            '--remote-source', 'source-id=test_source_id',
+            '--volume', self.new_snapshot.volume_id,
+        ]
+        ref_dict = {'source-name': 'test_source_name',
+                    'source-id': 'test_source_id'}
+        verifylist = [
+            ('remote_source', ref_dict),
+            ('volume', self.new_snapshot.volume_id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.snapshots_mock.manage.assert_called_with(
+            volume_id=self.new_snapshot.volume_id,
+            ref=ref_dict,
+            name=None,
+            description=None,
+            metadata=None,
+        )
+        self.snapshots_mock.create.assert_not_called()
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, data)
+
 
 class TestSnapshotDelete(TestSnapshot):
 
@@ -138,7 +196,7 @@ class TestSnapshotDelete(TestSnapshot):
         self.snapshots_mock.delete.return_value = None
 
         # Get the command object to mock
-        self.cmd = snapshot.DeleteSnapshot(self.app, None)
+        self.cmd = volume_snapshot.DeleteVolumeSnapshot(self.app, None)
 
     def test_snapshot_delete(self):
         arglist = [
@@ -152,7 +210,24 @@ class TestSnapshotDelete(TestSnapshot):
         result = self.cmd.take_action(parsed_args)
 
         self.snapshots_mock.delete.assert_called_with(
-            self.snapshots[0].id)
+            self.snapshots[0].id, False)
+        self.assertIsNone(result)
+
+    def test_snapshot_delete_with_force(self):
+        arglist = [
+            '--force',
+            self.snapshots[0].id
+        ]
+        verifylist = [
+            ('force', True),
+            ("snapshots", [self.snapshots[0].id])
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+
+        self.snapshots_mock.delete.assert_called_with(
+            self.snapshots[0].id, True)
         self.assertIsNone(result)
 
     def test_delete_multiple_snapshots(self):
@@ -168,7 +243,7 @@ class TestSnapshotDelete(TestSnapshot):
 
         calls = []
         for s in self.snapshots:
-            calls.append(call(s.id))
+            calls.append(call(s.id, False))
         self.snapshots_mock.delete.assert_has_calls(calls)
         self.assertIsNone(result)
 
@@ -199,13 +274,14 @@ class TestSnapshotDelete(TestSnapshot):
 
             self.assertEqual(2, find_mock.call_count)
             self.snapshots_mock.delete.assert_called_once_with(
-                self.snapshots[0].id
+                self.snapshots[0].id, False
             )
 
 
 class TestSnapshotList(TestSnapshot):
 
     volume = volume_fakes.FakeVolume.create_one_volume()
+    project = project_fakes.FakeProject.create_one_project()
     snapshots = volume_fakes.FakeSnapshot.create_snapshots(
         attrs={'volume_id': volume.name}, count=3)
 
@@ -248,22 +324,32 @@ class TestSnapshotList(TestSnapshot):
         super(TestSnapshotList, self).setUp()
 
         self.volumes_mock.list.return_value = [self.volume]
+        self.volumes_mock.get.return_value = self.volume
+        self.project_mock.get.return_value = self.project
         self.snapshots_mock.list.return_value = self.snapshots
         # Get the command to test
-        self.cmd = snapshot.ListSnapshot(self.app, None)
+        self.cmd = volume_snapshot.ListVolumeSnapshot(self.app, None)
 
     def test_snapshot_list_without_options(self):
         arglist = []
         verifylist = [
             ('all_projects', False),
-            ("long", False)
+            ('long', False)
         ]
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
         columns, data = self.cmd.take_action(parsed_args)
 
         self.snapshots_mock.list.assert_called_once_with(
-            limit=None, marker=None, search_opts={'all_tenants': False})
+            limit=None, marker=None,
+            search_opts={
+                'all_tenants': False,
+                'name': None,
+                'status': None,
+                'project_id': None,
+                'volume_id': None
+            }
+        )
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, list(data))
 
@@ -271,11 +357,13 @@ class TestSnapshotList(TestSnapshot):
         arglist = [
             "--long",
             "--limit", "2",
+            "--project", self.project.id,
             "--marker", self.snapshots[0].id,
         ]
         verifylist = [
             ("long", True),
             ("limit", 2),
+            ("project", self.project.id),
             ("marker", self.snapshots[0].id),
             ('all_projects', False),
         ]
@@ -286,7 +374,13 @@ class TestSnapshotList(TestSnapshot):
         self.snapshots_mock.list.assert_called_once_with(
             limit=2,
             marker=self.snapshots[0].id,
-            search_opts={'all_tenants': False}
+            search_opts={
+                'all_tenants': True,
+                'project_id': self.project.id,
+                'name': None,
+                'status': None,
+                'volume_id': None
+            }
         )
         self.assertEqual(self.columns_long, columns)
         self.assertEqual(self.data_long, list(data))
@@ -304,7 +398,93 @@ class TestSnapshotList(TestSnapshot):
         columns, data = self.cmd.take_action(parsed_args)
 
         self.snapshots_mock.list.assert_called_once_with(
-            limit=None, marker=None, search_opts={'all_tenants': True})
+            limit=None, marker=None,
+            search_opts={
+                'all_tenants': True,
+                'name': None,
+                'status': None,
+                'project_id': None,
+                'volume_id': None
+            }
+        )
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, list(data))
+
+    def test_snapshot_list_name_option(self):
+        arglist = [
+            '--name', self.snapshots[0].name,
+        ]
+        verifylist = [
+            ('all_projects', False),
+            ('long', False),
+            ('name', self.snapshots[0].name),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.snapshots_mock.list.assert_called_once_with(
+            limit=None, marker=None,
+            search_opts={
+                'all_tenants': False,
+                'name': self.snapshots[0].name,
+                'status': None,
+                'project_id': None,
+                'volume_id': None
+            }
+        )
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, list(data))
+
+    def test_snapshot_list_status_option(self):
+        arglist = [
+            '--status', self.snapshots[0].status,
+        ]
+        verifylist = [
+            ('all_projects', False),
+            ('long', False),
+            ('status', self.snapshots[0].status),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.snapshots_mock.list.assert_called_once_with(
+            limit=None, marker=None,
+            search_opts={
+                'all_tenants': False,
+                'name': None,
+                'status': self.snapshots[0].status,
+                'project_id': None,
+                'volume_id': None
+            }
+        )
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, list(data))
+
+    def test_snapshot_list_volumeid_option(self):
+        arglist = [
+            '--volume', self.volume.id,
+        ]
+        verifylist = [
+            ('all_projects', False),
+            ('long', False),
+            ('volume', self.volume.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.snapshots_mock.list.assert_called_once_with(
+            limit=None, marker=None,
+            search_opts={
+                'all_tenants': False,
+                'name': None,
+                'status': None,
+                'project_id': None,
+                'volume_id': self.volume.id
+            }
+        )
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, list(data))
 
@@ -330,9 +510,25 @@ class TestSnapshotSet(TestSnapshot):
         self.snapshots_mock.set_metadata.return_value = None
         self.snapshots_mock.update.return_value = None
         # Get the command object to mock
-        self.cmd = snapshot.SetSnapshot(self.app, None)
+        self.cmd = volume_snapshot.SetVolumeSnapshot(self.app, None)
 
-    def test_snapshot_set(self):
+    def test_snapshot_set_no_option(self):
+        arglist = [
+            self.snapshot.id,
+        ]
+        verifylist = [
+            ("snapshot", self.snapshot.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.snapshots_mock.get.assert_called_once_with(parsed_args.snapshot)
+        self.assertNotCalled(self.snapshots_mock.reset_state)
+        self.assertNotCalled(self.snapshots_mock.update)
+        self.assertNotCalled(self.snapshots_mock.set_metadata)
+        self.assertIsNone(result)
+
+    def test_snapshot_set_name_and_property(self):
         arglist = [
             "--name", "new_snapshot",
             "--property", "x=y",
@@ -359,6 +555,51 @@ class TestSnapshotSet(TestSnapshot):
         )
         self.assertIsNone(result)
 
+    def test_snapshot_set_with_no_property(self):
+        arglist = [
+            "--no-property",
+            self.snapshot.id,
+        ]
+        verifylist = [
+            ("no_property", True),
+            ("snapshot", self.snapshot.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.snapshots_mock.get.assert_called_once_with(parsed_args.snapshot)
+        self.assertNotCalled(self.snapshots_mock.reset_state)
+        self.assertNotCalled(self.snapshots_mock.update)
+        self.assertNotCalled(self.snapshots_mock.set_metadata)
+        self.snapshots_mock.delete_metadata.assert_called_with(
+            self.snapshot.id, ["foo"]
+        )
+        self.assertIsNone(result)
+
+    def test_snapshot_set_with_no_property_and_property(self):
+        arglist = [
+            "--no-property",
+            "--property", "foo_1=bar_1",
+            self.snapshot.id,
+        ]
+        verifylist = [
+            ("no_property", True),
+            ("property", {"foo_1": "bar_1"}),
+            ("snapshot", self.snapshot.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.snapshots_mock.get.assert_called_once_with(parsed_args.snapshot)
+        self.assertNotCalled(self.snapshots_mock.reset_state)
+        self.assertNotCalled(self.snapshots_mock.update)
+        self.snapshots_mock.delete_metadata.assert_called_with(
+            self.snapshot.id, ["foo"]
+        )
+        self.snapshots_mock.set_metadata.assert_called_once_with(
+            self.snapshot.id, {"foo_1": "bar_1"})
+        self.assertIsNone(result)
+
     def test_snapshot_set_state_to_error(self):
         arglist = [
             "--state", "error",
@@ -457,7 +698,7 @@ class TestSnapshotShow(TestSnapshot):
 
         self.snapshots_mock.get.return_value = self.snapshot
         # Get the command object to test
-        self.cmd = snapshot.ShowSnapshot(self.app, None)
+        self.cmd = volume_snapshot.ShowVolumeSnapshot(self.app, None)
 
     def test_snapshot_show(self):
         arglist = [
@@ -485,7 +726,7 @@ class TestSnapshotUnset(TestSnapshot):
         self.snapshots_mock.get.return_value = self.snapshot
         self.snapshots_mock.delete_metadata.return_value = None
         # Get the command object to mock
-        self.cmd = snapshot.UnsetSnapshot(self.app, None)
+        self.cmd = volume_snapshot.UnsetVolumeSnapshot(self.app, None)
 
     def test_snapshot_unset(self):
         arglist = [
